(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009, 2010)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Nnf
imports "../TermFun"
begin

datatype fm = And fm fm | Or fm fm | Imp fm fm | Iff fm fm | NOT fm | At nat
datatype_compat fm
setup \<open>Term_Fun.register_datatypes [@{type_name fm}]\<close>

function (sequential) nnf :: "fm \<Rightarrow> fm" where
  "nnf (At n) = At n"
| "nnf (And p q) = And (nnf p) (nnf q)"
| "nnf (Or p q) = Or (nnf p) (nnf q)"
| "nnf (Imp p q) = Or (nnf (NOT p)) (nnf q)"
| "nnf (Iff p q) = Or (And (nnf p) (nnf q)) (And (nnf (NOT p)) (nnf (NOT q)))"
| "nnf (NOT (And p q)) = Or (nnf (NOT p)) (nnf (NOT q))"
| "nnf (NOT (Or p q)) = And (nnf (NOT p)) (nnf (NOT q))"
| "nnf (NOT (Imp p q)) = And (nnf p) (nnf (NOT q))"
| "nnf (NOT (Iff p q)) = Or (And (nnf p) (nnf (NOT q))) (And (nnf (NOT p)) (nnf q))"
| "nnf (NOT (NOT p)) = nnf p"
| "nnf (NOT p) = NOT p"
  by pat_completeness auto

define_trs (no_sim) nnf.psimps

termination nnf
  apply (trs_reduce nnf)
  apply expand_trs_problem
  export_termination_problem "nnf.xml"
  apply (ceta_import_proof "proofs/nnf.proof.xml")
  apply eval
done

end
